#include <linux/spi/spidev.h>
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <stdint.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>

#include <string>
#include <thread>

#define SPI_DEVICE "/dev/spidev0.0"
#define SPI_TRANSFER_SIZE (4 * (1024 + 60))
#define SPI_BITS_PER_WORD 8
#define SPI_SPEED 25e6
#define SPI_MODE SPI_MODE_1
#define DELAY_BETWEEN_ITERATIONS_MS 50
#define THREAD_COUNT 3

int spi_transfer(const char *device, const uint8_t *write_buf, uint8_t *read_buf, uint32_t buf_len)
{
    int fd;
    int rc = -1;

    fd = open(device, O_RDWR);
    if (fd < 0)
    {
        printf("%s: failed to start SPI\r\n", __func__);
        return fd;
    }

    uint32_t mode = SPI_MODE;
    rc = ioctl(fd, SPI_IOC_WR_MODE32, &mode);
    if (rc < 0)
    {
        printf("%s: failed to set SPI write mode: %s\r\n", __func__, strerror(errno));
        close(fd);
        return rc;
    }

    rc = ioctl(fd, SPI_IOC_RD_MODE32, &mode);
    if (rc < 0)
    {
        // get error string from rc

        printf("%s: failed to set SPI read mode: %s\r\n", __func__, strerror(errno));
        close(fd);
        return rc;
    }

    // max buffer length is 4096 in spidev.c, but trying to write 4096 bytes in 32 bit mode
    // results in the ioctl never returning.
    do
    {
        uint32_t len = buf_len;

        if (len > 4096)
        {
            len = 4096;
        }
        printf("len: %d\r\n", len);
        struct spi_ioc_transfer transfer;
        memset(&transfer, 0, sizeof(transfer));
        transfer.tx_buf = (__u64)write_buf;
        transfer.rx_buf = (__u64)read_buf;
        transfer.len = len;
        transfer.speed_hz = SPI_SPEED;
        transfer.bits_per_word = SPI_BITS_PER_WORD;
        transfer.cs_change = 1;

        rc = ioctl(fd, SPI_IOC_MESSAGE(1), &transfer);
        if (rc < 0)
        {
            printf("%s: failed to start SPI transfer: %s\r\n", __func__, strerror(rc));
            close(fd);
            return rc;
        }
        if (write_buf)
        {
            write_buf += len;
        }
        if (read_buf)
        {
            read_buf += len;
        }
        buf_len -= len;
    } while (buf_len > 0);

    rc = close(fd);
    if (rc < 0)
    {
        printf("%s: failed to stop SPI\r\n", __func__);
    }

    return rc;
}

int i2c_read(uint8_t i2c_index, uint8_t address, uint8_t *data, int size)
{
    char filename[32];
    snprintf(filename, sizeof(filename), "/dev/i2c-%d", i2c_index);
    int fd = open(filename, O_RDWR);

    if (fd < 0)
    {
        return -1;
    }

    if (ioctl(fd, I2C_SLAVE, address) < 0)
    {
        close(fd);
        return -1;
    }

    if (read(fd, data, size) != size)
    {
        close(fd);
        return -1;
    }

    close(fd);
    return size;
}

void *spi_loop(const std::string &dev)
{
    float data[SPI_TRANSFER_SIZE];
    const char *device = dev.c_str();

    while (1)
    {
        // uint8_t buf[4];
        // spi_transfer(device, buf, NULL, sizeof(buf));
        spi_transfer(device, NULL, reinterpret_cast<uint8_t *>(data), SPI_TRANSFER_SIZE);

        std::this_thread::sleep_for(std::chrono::milliseconds(DELAY_BETWEEN_ITERATIONS_MS));
    }
    return NULL;
}

int main(void)
{
    auto spi_thread1 = std::thread(spi_loop, SPI_DEVICE);
    spi_thread1.join();

    return 0;
}